from pydantic import BaseModel
from nonebot.plugin import get_plugin_config


class Config(BaseModel):
    github_proxy: str = 'https://gh-proxy.com/'

    llm_helper_api_key: str
    llm_helper_model: str
    llm_helper_base_url: str

    llm_helper_timeout: int = 600 # 请求 LLM 的等待时间
    llm_helper_max_retries: int = 3 # 请求 Github 仓库的最大重试次数
    llm_helper_retry_delay: int = 10 # 请求 Github 仓库的重试间隔时间


config = get_plugin_config(Config)



import json
from pathlib import Path
from dataclasses import asdict

from nonebot import require
require('nonebot_plugin_localstore')
from nonebot_plugin_localstore import get_plugin_data_file

from .plugin import Plugin, Command

prompt_path = Path(__file__).parent / 'prompt.txt'
data_file_path = get_plugin_data_file('helper.json')


prompt = prompt_path.read_text(encoding='Utf-8')


def default_serializer(dump_object):
    if hasattr(dump_object, 'as_posix'):
        return dump_object.as_posix()
    return str(dump_object)


def save_helper(plugins: set[Plugin]) -> None:
    data = {
        plugin.name: [asdict(command) for command in plugin.helper]
        for plugin in plugins if plugin.helper
    }
    data_file_path.write_text(
        json.dumps(data, ensure_ascii=False, indent=2, default=default_serializer),
        encoding='utf-8'
    )


def load_helper(plguins: set[Plugin]) -> None:
    if not data_file_path.exists():
        return None
    data = json.loads(data_file_path.read_text(encoding='Utf-8'))
    for plugin in plguins:
        if plugin.name in data:
            plugin.helper = [Command(**command) for command in data[plugin.name]]



import re
import json
import asyncio
from pathlib import Path
from nonebot.log import logger

from .plugin import Plugin, Command
from .network import request_openai, fetch_github
from .data import save_helper
from .config import config
from .data import prompt


def list_dir(base_path: Path) -> str:
    # 列出所有目录，使用平铺的格式，并且不需要按层级，每个文件都是插件目录的相对路径
    children = []
    for path in base_path.rglob('*'):
        text_path = str(path.relative_to(base_path))
        if not '__pycache__' in text_path:
            children.append(text_path)
    return '\n'.join(children)


def read_file(base_path: Path, path: str) -> str:
    if '..' in path:
        return '你想要读取的文件路径不合法，请重新输入。'
    file_path = base_path / Path(path)
    return file_path.read_text(encoding='Utf-8')


async def generate_plugin_help(plugin: Plugin) -> bool:
    retry_count = 0
    readme = None
    if plugin.meta.homepage:
        # 先得到插件的README.md
        repo_name = plugin.meta.homepage.split('github.com/')[1]
        response = await fetch_github(f'https://raw.githubusercontent.com/{repo_name}/master/README.md')
        if response is not None:
            readme = response.text
            logger.debug(f'成功获取 {plugin.name} 的 README，文件内容：{readme}')
    readme_input = '以下为此项目的 README 文件内容：\n\n' + readme if readme else '暂时无法获取到 README 文件，请你通过把阅读代码来生成插件使用方法和说明文档。'
    messages = [
        {'role': 'system', 'content': prompt},
        {'role': 'user', 'content': readme_input}
    ]
    while True:
        if retry_count > config.llm_helper_max_retries:
            logger.error(f'{plugin.name} 的 LLM 响应失败！重试次数过多，跳过！')
            return False
        response = await request_openai(messages)
        if response is None:
            retry_count += 1
            logger.warning(f'{plugin.name} 的 LLM 响应失败！重试……')
            continue
        logger.debug(f'{plugin.name} 的 LLM 响应：{response}')
        messages.append({'role': 'assistant', 'content': response})
        response = response.split('</think>')[-1].strip()
        if result_match := re.match(r'<result>(.*)</result>', response, re.DOTALL):
            string_result = result_match.group(1)
            logger.debug(f'{plugin.name} 的 LLM 生成帮助结果：{string_result}')
            try:
                result = json.loads(string_result.replace('\'', '"'))
                plugin.helper = [Command(**command) for command in result]
            except json.JSONDecodeError:
                retry_count += 1
                logger.error(f'{plugin.name} 的 LLM 生成帮助结果格式错误！重试……')
                messages.append({'role': 'user', 'content': '你的输出的 Json 格式错误，请重新检查并输出。请注意要将格式正确的 Json 输出在 <result> 标签内，否则我无法解析。'})
                continue
            logger.success(f'{plugin.name} 的 LLM 生成帮助成功！')
            logger.debug(f'{plugin.name} 的 LLM 生成帮助结果：{result}')
            return True
        if re.match(r'<list_dir>(.*)</list_dir>', response, re.DOTALL):
            list_dir_content = list_dir(plugin.path)
            messages.append({'role': 'user', 'content': '以下为此项目的目录结构：\n\n' + list_dir_content})
            continue
        if read_file_match := re.match(r'<read_file>(.*)</read_file>', response, re.DOTALL):
            read_file_content = read_file(plugin.path, read_file_match.group(1))
            messages.append({'role': 'user', 'content': '以下为你想要读取的代码文件内容：\n\n' + read_file_content})
            continue
        retry_count += 1
        logger.warning(f'{plugin.name} 的 LLM 响应格式错误！重试……')
        messages.append({'role': 'user', 'content': '请你输出一个工具，否则我无法与你继续对话。'})


async def generate_plugins_help(plugins: set[Plugin]) -> None:
    tasks = []
    logger.info('正在生成插件帮助……')
    for plugin in plugins:
        if plugin.helper:
            logger.debug(f'{plugin.name} 已存在帮助，直接使用缓存。')
            continue
        if plugin.meta is None or plugin.path is None:
            logger.info(f'{plugin.name} 未找到元数据与路径，跳过！')
            continue
        tasks.append(asyncio.create_task(generate_plugin_help(plugin)))
    success_count = sum(await asyncio.gather(*tasks))
    logger.success(f'插件帮助生成完毕！成功生成 {success_count} 个插件的帮助，失败 {len(tasks) - success_count} 个插件。')
    save_helper(plugins)



from typing import AsyncGenerator, Optional

from .plugin import Plugin


async def generator_to_string(generator: AsyncGenerator[str, None]) -> str:
    result = []
    async for line in generator:
        result.append(line)
    return '\n'.join(result)


def get_plugin_by_name(plugins: set[Plugin], plugin_name: str) -> Optional[Plugin]:
    return next((plugin for plugin in plugins if plugin.name == plugin_name), None)



from pathlib import Path
from dataclasses import dataclass, field
from typing import Optional
from nonebot.plugin import get_loaded_plugins
from nonebot.plugin import PluginMetadata


@dataclass
class Param:
    value: str
    optional: bool
    description: str


@dataclass
class Command:
    name: str
    description: str
    alias: list[str] = field(default_factory=list)
    params: dict[str, Param] = field(default_factory=dict)

    def __post_init__(self) -> None:
        self.params = {
            key: Param(**param) if not isinstance(param, Param) else param
            for key, param in self.params.items()
        }


@dataclass
class Plugin:
    id: str
    name: Optional[str] = None
    path: Optional[Path] = None
    meta: Optional[PluginMetadata] = None
    helper: list[Command] = field(default_factory=list)

    def __hash__(self):
        return hash(self.id)

    def __post_init__(self) -> None:
        self.helper = [
            Command(**command) if not isinstance(command, Command) else command
            for command in self.helper
        ]


def search_plugins() -> set[Plugin]:
    plugins = set()
    for plugin in get_loaded_plugins():
        plugin_meta: PluginMetadata | None = getattr(plugin.module, '__plugin_meta__', None)
        if (not plugin_meta) or plugin_meta.type == 'library':
            continue
        if plugin.module.__file__:
            path = Path(plugin.module.__file__)
            plugin_name = plugin_meta.name.replace(' ', '-')
            plugins.add(Plugin(id=plugin.module_name, name=plugin_name, meta=plugin_meta, path=path.parent))
            continue
        plugins.add(Plugin(id=plugin.module_name))
    return plugins



import asyncio
from nonebot import get_driver, require
from nonebot.log import logger
from nonebot.plugin import PluginMetadata
require('nonebot_plugin_alconna')
from nonebot_plugin_alconna import Command, Arparma

from .config import Config
from .generate import generate_plugins_help, generate_plugin_help
from .plugin import Plugin, search_plugins
from .data import load_helper
from .utils import get_plugin_by_name, generator_to_string

__plugin_meta__ = PluginMetadata(
    name='LLM-Helper',
    description='一个支持多平台的帮助插件，基于 LLM 解析 Github 仓库或代码。',
    usage='可以通过命令 /help 或 /llm-help 进行查询。',
    homepage='https://github.com/Lonely-Sails/nonebot-plugin-llm-helper',
    type='application',
    config=Config,
)

plugins: set[Plugin] = set()
adapter = get_driver()

alconna_command = Command('llm-help [plugin:str]')
alconna_command.alias('help')
alconna_command.subcommand('list')
alconna_command.subcommand('regenerate regen')
alconna_command.subcommand('search [keyword:str]')
matcher = alconna_command.build(use_cmd_start=True)

async def list_plugins():
    yield '所有插件列表：'
    for plugin in plugins:
        if plugin.meta:
            yield f'\n  {("☑️", "✅")[bool(plugin.helper)]} {plugin.name}：'
            yield f'    - {plugin.meta.description}'
            continue
        yield f'\n  ☑️ {plugin.name}：未找到元数据。'


async def list_plugin_commands(plugin: Plugin):
    if (not plugin.name) or (plugin is None):
        yield f'插件 {plugin.name} 不存在，请检查你的输入是否正确。'
        return
    yield f'插件 {plugin.name} 的命令如下：'
    for command in plugin.helper:
        yield f'\n  {command.name}：{command.description}'
        if command.alias:
            yield f'    - 别名：{" ".join(command.alias)}'
        yield f'    - 用法：{command.name} {" ".join([f"[{param_name}]" if param_info.optional else f"<{param_name}>" for param_name, param_info in command.params.items()])}'
        for param_name, param_info in command.params.items():
            yield f'      {"*" if param_info.optional else "+"} {param_name}：{param_info.description}'
            yield f'        参数值描述：{param_info.value}'

    
async def regenerate_plugin(plugin: Plugin):
    if await generate_plugin_help(plugin):
        yield f'插件 {plugin.name} 的帮助已重新生成。'
        return
    yield f'插件 {plugin.name} 的帮助重新生成失败，请稍后再试。'


async def search_commands(keyword: str):
    count = 0
    yield f'搜索关键词 {keyword} 的结果：\n'
    keyword = keyword.lower()
    for plugin in plugins:
        if not plugin.name:
            continue
        if keyword in plugin.name.lower():
            yield f'  插件 {plugin.name} -> {plugin.meta.description if plugin.meta else "暂无简介"}'
            count += 1
        for command in plugin.helper:
            search_string = command.name + command.description + ' '.join(command.alias)
            if keyword in search_string.lower():
                yield f'  命令 {plugin.name}：{command.name} -> {command.description}'
                count += 1
    if count == 0:
        yield f'  暂未找到与之相关的命令或插件。'
        return
    yield f'\n共找到 {count} 个结果。'


@adapter.on_startup
async def _():
    global plugins
    plugins = search_plugins()
    all_count = len(plugins)
    none_meta_count = sum(plugin.meta is None for plugin in plugins)
    logger.info(f'搜索插件完毕：共 {all_count} 个有效的用户插件，{none_meta_count} 个插件未找到元数据与路径已忽略。')
    load_helper(plugins)
    asyncio.create_task(generate_plugins_help(plugins))


@matcher.handle()
async def _(result: Arparma):
    if 'list' in result.subcommands:
        await matcher.finish(await generator_to_string(list_plugins()))
    elif search_result := result.subcommands.get('search', None):
        keyword = search_result.args.get('keyword', None)
        if keyword is None:
            await matcher.finish(f'请输入搜索的关键词。')
        await matcher.finish(await generator_to_string(search_commands(keyword)))
    plugin_name = result.query[str]('main_args.plugin')
    if plugin_name is None:
        await matcher.finish(f'请输入插件名称。')
    plugin = get_plugin_by_name(plugins, plugin_name)
    if not plugin:
        await matcher.finish(f'插件 {plugin_name} 不存在，请检查你的输入是否正确。')
    if 'regenerate' in result.subcommands:
        await matcher.send(f'正在重新生成插件 {plugin.name} 的帮助，请耐心等待……')
        await matcher.finish(await generator_to_string(regenerate_plugin(plugin)))
    await matcher.finish(await generator_to_string(list_plugin_commands(plugin)))


import asyncio
from typing import Optional
from httpx import AsyncClient, Response
from nonebot.log import logger

from .config import config

client = AsyncClient(timeout=config.llm_helper_timeout)
llm_headers = {'Authorization': f'Bearer {config.llm_helper_api_key}', 'Content-Type': 'application/json'}


async def fetch_github(url: str, count: int = 0) -> Optional[Response]:
    try:
        response = await client.get(config.github_proxy + url, timeout=30)
        response.raise_for_status()
        return response
    except Exception as error:
        logger.warning(f'第 {count + 1} 次请求 {url} 失败：{error}')
        await asyncio.sleep(config.llm_helper_retry_delay)
        if count < config.llm_helper_max_retries:
            return await fetch_github(url, count + 1)
        return None


async def request_openai(messages: list[dict]) -> Optional[str]:
    json = {
        'model': config.llm_helper_model,
        'messages': messages,
    }
    try:
        response = await client.post(
            f'{config.llm_helper_base_url}/chat/completions',
            headers=llm_headers, json=json,
        )
        response.raise_for_status()
        response_json = response.json()
        return response_json['choices'][0]['message']['content']
    except Exception as error:
        logger.warning(f'OpenAI 请求失败：{error}')
        return None